extern "C" {
  #include "../aiger/aiger.h"
  #include "../aiger/simpaig.h"
}

#include <unistd.h>
#include <sys/types.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <assert.h>
#include <signal.h>
#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <set>
#include <algorithm>
#include <cctype>
#include <ctime>
#include "aigcertify.hh"

using namespace std;

aiger *model;
aiger *certificate;
aiger *con1;
aiger *con2;
aiger *con3;

unsigned m_num_inputs; /* model & certificate should have the same number of inputs */
unsigned m_num_latches; /* model & certificate should have the same number of latches*/
unsigned m_num_ands;
unsigned m_total_num_literals;

unsigned c_num_inputs; /* model & certificate should have the same number of inputs */
unsigned c_num_latches; /* model & certificate should have the same number of latches*/
unsigned c_num_ands;
unsigned c_total_num_literals;

unsigned neg(unsigned l){
  if (l%2 == 1) return l-1;
  return l+1;
}

bool m_is_input(unsigned l){
  unsigned index = l/2;
  return l<=m_num_inputs;
}

bool m_is_latch(unsigned index) {
  unsigned offset = index/2;
  return offset>m_num_inputs
         && offset<=(m_num_inputs + m_num_latches);
}

bool m_is_and(unsigned index){
  unsigned offset = index/2;
  return offset>m_num_inputs + m_num_latches
         && offset <= m_total_num_literals;
}

bool c_is_input(unsigned l){
  unsigned index = l/2;
  return index<=c_num_inputs;
}

bool c_is_latch(unsigned index) {
  unsigned offset = index/2;
  return offset>c_num_inputs
         && offset<=(c_num_inputs + c_num_latches);
}

bool c_is_and(unsigned index){
  unsigned offset = index/2;
  return offset>c_num_inputs + c_num_latches
         && offset <= c_total_num_literals;
}

unsigned mappingModelIndex2Con2(unsigned l){
  if(l==0 || l==1) return l;
  if(m_is_input(l)){
    return l;
  }else if(m_is_latch(l)){
    return l;
  }else if(m_is_and(l)){
    return l + c_num_ands *2;
  }
}

unsigned map_original(unsigned l){
  if(l==0 || l==1) return l;
  if(c_is_input(l)){
    return l;
  }else if(c_is_latch(l)){
    return l + m_num_inputs * 2;
  }else if(c_is_and(l)){
    return l + m_num_inputs * 2;
  }
}

unsigned map_to_next(unsigned l){
  if(l==0 || l==1) return l;
  if(c_is_input(l)){
    return l + m_num_inputs * 2;
  }else if(c_is_latch(l)){
    unsigned idx = l/2;
    aiger_symbol* s = aiger_is_latch(certificate, l);
    unsigned new_next = map_original(s->next);
    return l%2? neg(new_next) : new_next;
  }else if(c_is_and(l)){
    return l +  m_num_inputs * 2 + c_num_ands * 2;
  }
}

int 
main (int argc, char *argv[])
{
  /* output file */
  std::string output_file("-o");

  const char* output_name1;
  const char* output_name2;
  const char* output_name3;
  const char* model_name;
  const char* certificate_name;

  for (int i = 1; i < argc - 1; ++i) {
    if (output_file == argv[i]){
      output_name1 = argv[++i];
      output_name2 = argv[++i];
      output_name3 = argv[++i];
    }
  }

  assert(argc>=2);

  model_name = argv[argc-2];
  certificate_name = argv[argc-1];
  cout<<"Reading model file: "<<model_name<<endl;
  cout<<"Reading certificate file: "<<certificate_name<<endl;

  model = aiger_init();
  certificate = aiger_init();
  con1 = aiger_init();
  con2 = aiger_init();
  con3 = aiger_init();

  aiger_open_and_read_from_file(model, model_name);
  aiger_open_and_read_from_file(certificate, certificate_name);

  m_num_inputs = model->num_inputs;
  m_num_latches = model->num_latches;
  m_num_ands = model->num_ands;
  m_total_num_literals = m_num_inputs + m_num_latches + m_num_ands;

  aiger_symbol* m_latches = model->latches;
  aiger_and* m_ands = model->ands;

  c_num_inputs = certificate->num_inputs;
  c_num_latches = certificate->num_latches;
  c_num_ands = certificate->num_ands;
  c_total_num_literals = c_num_inputs + c_num_latches + c_num_ands;
  aiger_and* c_ands = certificate->ands;

  //assert(c_num_ands == m_num_ands);

  /* First condition: I & !C*/
  /* adding original inputs */
  for(unsigned i=0; i<m_num_inputs; i++){
    aiger_add_input(con1, (i+1)*2, "");
  }

  /* adding original latches as inputs */
  for(unsigned i=0; i<m_num_latches; i++){
    aiger_symbol* latch = m_latches + i;
    aiger_add_input(con1, latch->lit, "");
  }

  /* copy the certificate circuit */
  cout<<"c_num_ands:"<<c_num_ands<<endl;
  for(unsigned i=0; i<c_num_ands; i++){
    aiger_and* a = c_ands + i;
    aiger_add_and(con1, a->lhs, a->rhs0, a->rhs1);
  }

  unsigned current_index = m_total_num_literals*2 + 2;
  /* I: */
  vector<unsigned>* initLatches = new vector<unsigned>;
  for(unsigned i=0; i<model->num_latches; i++){
    aiger_symbol* latch = m_latches + i;
    unsigned l = latch->lit;
    unsigned reset = latch->reset;
    aiger_add_and(con1, current_index, l, neg(reset));
    current_index += 2;
    aiger_add_and(con1, current_index, neg(l), reset);
    current_index += 2;
    aiger_add_and(con1, current_index, current_index-3, current_index-1);
    initLatches->push_back(current_index);
    current_index += 2;
  }

  unsigned prev = initLatches->at(0);
  for(unsigned i=1;i<initLatches->size(); i++){
    aiger_add_and(con1, current_index, prev, initLatches->at(i));
    prev = current_index;
    current_index += 2;
  }

  unsigned initialStates = current_index-2;

  /* !C */
  aiger_symbol* c_output = certificate->outputs;
  aiger_symbol* m_output = model->outputs;
  unsigned b = m_output->lit;
  unsigned c = neg(c_output->lit);
  aiger_add_and(con1, current_index, initialStates, neg(c));
  aiger_add_output(con1, current_index, "");

  if(output_name1){
    aiger_open_and_write_to_file (con1, output_name1);
  }else{
    aiger_write_to_file (con1, aiger_ascii_mode, stdout);
  }

  /* COND2: C & B */
  /* adding original inputs */
  for(unsigned i=0; i<m_num_inputs; i++){
    aiger_add_input(con2, (i+1)*2, "");
  }

  /* adding original latches as inputs */
  for(unsigned i=0; i<m_num_latches; i++){
    aiger_symbol* latch = m_latches + i;
    aiger_add_input(con2, latch->lit, "");
  }

  /* copy the certificate circuit */
  for(unsigned i=0; i<c_num_ands; i++){
    aiger_and* a = c_ands + i;
    aiger_add_and(con2, a->lhs, a->rhs0, a->rhs1);
    current_index = a->lhs;
  }

  current_index += 2;
  aiger_add_and(con2, current_index, c, b);
  aiger_add_output(con2, current_index, "");
  if(output_name2){
    aiger_open_and_write_to_file (con2, output_name2);
  }else{
    aiger_write_to_file (con2, aiger_ascii_mode, stdout);
  }

  /* COND3: C & !C' */
  /* adding original inputs */
  for(unsigned i=0; i<m_num_inputs; i++){
    aiger_add_input(con3, (i+1)*2, "");
  }

  /* adding another set of inputs */
  for(unsigned i=0; i<m_num_inputs; i++){
    aiger_add_input(con3, m_num_inputs * 2 + (i+1)*2, "");
  }

  /* adding original latches as inputs */
  for(unsigned i=0; i<m_num_latches; i++){
    aiger_symbol* latch = m_latches + i;
    aiger_add_input(con3, m_num_inputs * 2 + latch->lit, "");
  }

  /* copy the certificate circuit */
  for(unsigned i=0; i<c_num_ands; i++){
    aiger_and* a = c_ands + i;
    aiger_add_and(con3, map_original(a->lhs), map_original(a->rhs0), map_original(a->rhs1));
  }

  for(unsigned i=0; i<c_num_ands; i++) {
    aiger_and* a = c_ands + i;
    unsigned lhs = map_to_next(a->lhs);
    unsigned rhs0 = map_to_next(a->rhs0);
    unsigned rhs1 = map_to_next(a->rhs1);

    aiger_add_and(con3, lhs, rhs0, rhs1);
    if(i == c_num_ands-1) current_index = lhs + 2;
  }

  unsigned cp = map_to_next(c);
  aiger_add_and(con3, current_index, map_original(c), neg(cp));

  aiger_add_output(con3, current_index, "");
  if(output_name3) {
    aiger_open_and_write_to_file (con3, output_name3);
  }else{
    aiger_write_to_file(con3, aiger_ascii_mode, stdout);
  }

  return 0;
}

