#!/usr/bin/env python3

import re
import sys
import os
import signal
from enum import Enum
from subprocess import Popen, PIPE
from pyparsing import nestedExpr
import pyparsing as pp
import re
import time
import uuid


def generate_smv_file(num_bits, m, k):
    f = open(sys.path[0] + "/temp_smv/q_"+str(num_bits)+".smv", "w")
    f.write("MODULE main\n")
    f.write(" VAR\n")
    for i in range(num_bits):
        f.write("    " + "b" + str(i) + ": boolean;\n")
    f.write(" DEFINE\n")
    f.write("   bad:= ")

    for i in range(m, num_bits-1):
        f.write("b"+str(i)+ " | ")
    f.write("b"+str(num_bits-1)+";\n")
    f.write(" ASSIGN\n")
    for i in range(num_bits):
        if i==0:
            f.write("  init(b0) := 1;\n")
        else:
            f.write("  init(b"+str(i)+") := 0;\n")

    for i in range(0, m+2):

        f.write("   next(b"+str(i) + ") := \n")
        f.write("    case\n")
        for j in range(0,num_bits-k-1):
            f.write("b"+str(j)+" & ")
        for j in range(num_bits-k-1, num_bits-1):
            f.write("!b"+str(j)+" & ")
        if i==0:
            f.write("!b"+str(num_bits-1)+": 1;\n")
        else:
            f.write("!b"+str(num_bits-1)+": 0;\n")
        for j in range(0, num_bits-1):
            f.write("b"+str(j)+" & ")
        f.write("b"+str(num_bits-1)+": 0;\n")
        if i==0:
            f.write("1: 1;\n")
        else:
            f.write("1: b" + str(i-1)+";\n")
        f.write("esac;\n")
    for i in range(m+2,num_bits):
        f.write("   next(b"+str(i) + ") := \n")
        f.write("    case\n")
        for j in range(0,num_bits-k-1):
            f.write("b"+str(j)+" & ")
        for j in range(num_bits-k-1, num_bits-1):
            f.write("!b"+str(j)+" & ")
        f.write("!b"+str(num_bits-1)+": 0;\n")

        for j in range(0,i):
            f.write("b"+str(j)+" & ")
        for j in range(i,num_bits-1):
            f.write("!b"+str(j)+" & ")
        f.write("!b"+str(num_bits-1)+": 1 ;\n")
        for j in range(0,num_bits-1):
            f.write("b"+str(j) + "&")
        f.write("b"+str(num_bits-1)+": 0;\n")
        f.write("1: b"+str(i-1) +";\n")
        f.write("       esac;\n")

    f.write("SPEC AG !bad");
    f.close()
    return f
    
def generator(bits, modulo, bad):
    exe = sys.path[0] + "/generators/generator -r -e"
    f = sys.path[0] + "/temp_smv/"+str(bits)+"m"+str(modulo)+"b"+str(bad)
    filename = f+ ".smv"
    
    cmd = exe + " counter "+str(bits) +" " + str(modulo) +" " + str(bad)     
    print("cmd", cmd)
    
    signal.signal(signal.SIGPIPE, signal.SIG_IGN)
    p=Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]
    output = output.splitlines()
    filew = open(filename, "w")
    for item in output:
        filew.write("%s\n" % item.decode('utf8'))
    filew.close
    return f

def smv_to_aig(f):
    print("converting smv to aig")
    exe = sys.path[0] + "/aiger/smvtoaig "
    #outfile = " q_"+str(num_bits)+".aag"
    newfilename = f.split("/")[-1]
    cmd = exe + " -s -a "  + f+".smv"
    print("cmd: ", cmd)

    signal.signal(signal.SIGPIPE, signal.SIG_IGN)
    p=Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]
    output = output.splitlines()

    #convert_to_format1(f)
    newfile = open(sys.path[0]+"/temp/"+newfilename+".aag", "w")
    for i,line in enumerate(output):
        line = line.decode('utf8')
        if i==0:
            words = line.split(" ")[:-1]
            newline = ""
            print(words)
            for j,w in enumerate(words):
                if j==4:
                    newline = newline + "1 "
                elif j==len(words)-1:
                    newline = newline + (w) + "\n"
                else:
                    newline = newline + (w + " ")
            newfile.write(newline)
        else:
            newfile.write(line+"\n")
    newfile.close()
    return newfile

def mcaiger(f, num_bits):
    print("calling Mcaiger...")
    exe = sys.path[0] + "/mcaiger/mcaiger -n "
    cmd = exe + f.name
    print("cmd: ", cmd)
    
    time1 = time.time()
    signal.signal(signal.SIGPIPE, signal.SIG_IGN)
    p=Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]
    
    time_taken = time.time()-time1
    output = output.splitlines()
    k = output[-1].decode('utf8').split("=")[-1]
    
    print("k: ", k)
    print("McAiger computation time: ", time_taken)
    return k

def aigcertify(f,k_val):
    filename = "temp/" + f.name.split("/")[-1]
    cmd  = "python3 " +  sys.path[0] + "/py-aigcertify.py " + " " + filename + " -kis -kwm"

    print("cmd: ", cmd)
    signal.signal(signal.SIGPIPE, signal.SIG_IGN)
    p=Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]
    output = output.splitlines()
    result = output[-2].decode('utf8')
    #assert(result == "SUCCESS")
    print(output[-1].decode('utf8'))

def aigcertify_mc(f,k_val):
    filename = "temp/" + f.name.split("/")[-1]
    cmd  = "python3 " +  sys.path[0] + "/py-aigcertify.py " + " " + filename + " -kis -nmc "+str(k_val)

    print("cmd: ", cmd)
    signal.signal(signal.SIGPIPE, signal.SIG_IGN)
    p=Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]
    output = output.splitlines()
    result = output[-2].decode('utf8')
    #assert(result == "SUCCESS")
    print(output[-1].decode('utf8'))


def iter_aigcertify():
    for i in [500]:
        for j in [(2**5)]:
            for a in range(j+1,j+400,2):
                f = generator(i,j, a)
                aig_file = smv_to_aig(f)
                mcaiger(aig_file, i)
                aigcertify(aig_file, i)

def faulty_test():
    for i in [50]:
        for j in [(2**5)]:
            for a in range(j+20,j+50,5):
                f = generator(i,j, a)
                aig_file = smv_to_aig(f)
                k=mcaiger(aig_file, i)
                aigcertify_mc(aig_file, int(k)-1)

############ main #############

#max_num_bits = int(sys.argv[1])
#property_m = int(sys.argv[2])
#inductive_k = int(sys.argv[3])

if(len(sys.argv)==5):
    #does not iterate
    f = generate_smv_file(max_num_bits, property_m, inductive_k)
    aig_file = smv_to_aig(f, max_num_bits)
    mcaiger(aig_file, max_num_bits)
    aigcertify(aig_file, max_num_bits)
else:
    #iter_aigcertify()
    faulty_test()
