#!/bin/sh

# You need 'smvtoaig', 'aigmove' and 'mcaiger' in your path to run this.

die () {
  echo "test.sh: error: $*" 2>&1 
  exit 1
}

[ -f generator ] || die "can not find './generator' (issue 'make' first)"

for required in ./../aiger/smvtoaig ./../aiger/aigmove ../mcaiger/mcaiger
do
  $required -h >/dev/null 2>&1
  [ $? = 0 ] || die "could not find '$required'"
done

if [ -d log ]
then
  cd log
  echo "cleaning up existing 'log' directory for test output"
  rm -f *.smv *.aig1 *.aig2 *.log *.err
  cd ..
else
  mkdir log || exit 1
  echo "generated new 'log' directory for test output"
fi

run () {
  expected=$1
  shift
  options="$*"
  cmd="./generator $options"
  echo -n "$cmd # expected '$expected'"
  base="`echo $options |sed -e 's, ,-,g' -e 's,--*,-,g'`"
  prefix=log/$base
  smv=$prefix.smv
  aig1=$prefix.aig1
  aig2=$prefix.aig2
  log=$prefix.log
  err=$prefix.err
  rm -f $log $err
  ./$cmd > $smv 2>>$err
  smvtoaig $smv $aig1 1>>$log 2>>$err
  aigmove $aig1 $aig2 1>>$log 2>>$err
  stalling=x"`echo $base|grep -- '-s-'`"
  if [ $stalling = x ]
  then
    nosimplepathoption=" -n"
  else
    nosimplepathoption=""
  fi
  mcaiger -v$nosimplepathoption $aig2 1>>$log 2>>$err
  status=$?
  if [ $status = $expected ]
  then
    echo " got '$status' OK"
  else
    echo " got '$status' FAILED"
    echo "run.sh: more information in '$prefix.{smv,aig1,aig2,log,err}'"
    exit 1
  fi
}

for reset in "" "-r "
do
  for stall in "" "-s "
  do
    for size in 1 2 4 8
    do
      options="$reset$stall$size"
      run 20 queue $options
    done # reset
  done # stall
done # size

for reset in "" "-r "
do
  for stall in "" "-s "
  do
    for exactmodulo in "" "-m "
    do
      for exactbad in "" "-b "
      do
	for bits in 1 2 3 4
	do
	  size=`echo 2 $bits^p|dc`
	  modulo=1
	  while [ $modulo -le $size ]
	  do
	    bad=0
	    while [ $bad -lt $size ]
	    do
	      options="$reset$stall$exactmodulo$exactbad"
	      options="$options$bits $modulo $bad"
	      if [ $bad -lt $modulo ]
	      then
		expected=10
	      else
		expected=20
	      fi
	      run $expected counter $options
	      bad=`expr $bad + 1`
	      [ $bad = $size ] && break
	      bad=`expr $bad + 1`
	      [ $bad = $size ] && break
	    done # bad
	  [ $modulo = $size ] && break
	  modulo=`expr $modulo \* 3`
	  [ $modulo -gt $size ] && modulo=$size
	  done # modulo
	done # bits
      done # exactbad
    done # exactmodulo
  done # stall
done # reset

echo "all tests succeeded (see intermediate and log files in 'log' directory)"
