extern "C" {
#include "../aiger/aiger.h"
#include "../aiger/simpaig.h"
}

#include <unistd.h>
#include <sys/types.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <assert.h>
#include <signal.h>
#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <set>
#include <algorithm>
#include <cctype>
#include <ctime>

using namespace std;

aiger *model;
aiger *kw_model;
aiger *con1;
aiger *con2;
aiger *con3;
aiger *con4;//new reset condition`

unsigned k;

unsigned m_num_inputs;
unsigned m_num_latches;
unsigned m_num_ands;
unsigned m_total_num_literals;

unsigned km_num_inputs;
unsigned km_num_latches;
unsigned km_num_ands;
unsigned km_total_num_literals;

unsigned neg(unsigned l){
  if (l%2 == 1) return l-1;
  return l+1;
}

bool m_is_input(unsigned index){
  unsigned offset = index/2;
  return offset <= m_num_inputs;
}

bool m_is_latch(unsigned index) {
  unsigned offset = index/2;
  return offset>m_num_inputs
         && offset<=(m_num_inputs + m_num_latches);
}

bool m_is_and(unsigned index){
  unsigned offset = index/2;
  return offset>m_num_inputs + m_num_latches
         && offset <= m_total_num_literals;
}

/* mapping original circuit indices to con2 */
unsigned map_kwm_to_con2(unsigned index){
  if(index == 0 || index == 1) return index;
  unsigned new_index;
  if(m_is_input(index)){
    new_index = index;
  }else if(m_is_latch(index)){
    new_index = index + (k-1)* m_num_inputs *2;
  }else if(m_is_and(index)){
    new_index = index + (k-1)* m_num_inputs *2 + (k-1)*m_num_latches*2 + k*2 + km_num_ands * 2;
  }
  return new_index;
}

int
main (int argc, char *argv[])
{
  /* output file */
  std::string output_file("-o");
  std::string k_val("-k");
  const char* output_name1;
  const char* output_name2;
  const char* output_name3;

  const char* model_name;
  const char* kwmodel_name;

  k=-1;

  for (int i = 1; i < argc - 1; ++i) {
    if(k_val == argv[i]){
      string s = argv[++i];
      k = atoi(s.c_str());
    }else if (output_file == argv[i]){
      output_name1 = argv[++i];
      output_name2 = argv[++i];
      output_name3 = argv[++i];
    }
  }
  assert(argc>=3);

  model_name = argv[argc-2];
  kwmodel_name = argv[argc-1];
  cout<<"Reading model file: "<<model_name<<endl;
  cout<<"Reading k-witness model file: "<<kwmodel_name<<endl;

  model = aiger_init();
  kw_model = aiger_init();
  con1 = aiger_init();
  con4 = aiger_init();
  con2 = aiger_init();
  con3 = aiger_init();

  aiger_open_and_read_from_file(model, model_name);
  aiger_open_and_read_from_file(kw_model, kwmodel_name);

  m_num_inputs = model->num_inputs;
  m_num_latches = model->num_latches;
  m_num_ands = model->num_ands;
  m_total_num_literals = m_num_inputs + m_num_latches + m_num_ands;

  km_num_inputs = kw_model->num_inputs;
  km_num_latches = kw_model->num_latches;
  km_num_ands = kw_model->num_ands;
  km_total_num_literals = km_num_inputs + km_num_latches + km_num_ands;

  aiger_symbol* m_latches = model->latches;
  aiger_and* m_ands = model->ands;

  aiger_symbol* km_latches = kw_model->latches;
  aiger_and* km_ands = kw_model->ands;

  unsigned prev;
  /* first condition: f(I,L) == f'(I,L) */
  for(unsigned i=0; i<km_num_inputs; i++){
    aiger_add_input(con1, (i+1)*2, "");
    aiger_add_input(con4, (i+1)*2, "");
  } 
  /* adding kw latches */
  for(unsigned i=0; i<km_num_latches; i++){
    aiger_add_input(con1, km_num_inputs*2+(i+1)*2, "");
    aiger_add_input(con4, km_num_inputs*2+(i+1)*2, "");
  }

  /* adding kw ands */
  for(unsigned i=0; i<km_num_ands; i++){
    aiger_and* current_and = km_ands + i;
    aiger_add_and(con1, current_and->lhs, current_and->rhs0, current_and->rhs1);
    aiger_add_and(con4, current_and->lhs, current_and->rhs0, current_and->rhs1);
  }
  unsigned ci;
  /* adding original circuit with mapping */
  for(unsigned i=0; i<m_num_ands; i++){
    aiger_and* current_and = m_ands + i;
    ci = map_kwm_to_con2(current_and->lhs);
    aiger_add_and(con1, ci, map_kwm_to_con2(current_and->rhs0), map_kwm_to_con2(current_and->rhs1));
    aiger_add_and(con4, ci, map_kwm_to_con2(current_and->rhs0), map_kwm_to_con2(current_and->rhs1));
  }
  /* f=f' */
  if(m_num_latches == 0){
    aiger_add_output(con1, 1, "");
    aiger_add_output(con4, 1, "");
  }else{
  vector<unsigned>* fis = new vector<unsigned>;
  cout<<"m_num_latches "<<m_num_latches<<endl;
  for(unsigned i=0; i<m_num_latches; i++){
    aiger_symbol* l = m_latches+ i;
    aiger_symbol* l_prime = km_latches  + (k-1)* m_num_inputs + i;
    unsigned f = map_kwm_to_con2(l->next);
    unsigned f_prime = l_prime->next;
    ci+=2;
    aiger_add_and(con1, ci, f, neg(f_prime));
    unsigned r = map_kwm_to_con2(l->reset);
    unsigned r_prime = l_prime->reset;
    aiger_add_and(con4, ci, r, neg(r_prime));
    unsigned a1 = ci+1;
    ci+=2;
    aiger_add_and(con1, ci, neg(f), f_prime);
    aiger_add_and(con4, ci, neg(r), r_prime);
    unsigned a2 = ci+1;
    ci+=2;
    aiger_add_and(con1, ci, a1, a2);
    aiger_add_and(con4, ci, a1, a2);

    fis->push_back(ci);
  }
  prev = fis->at(0);
  for(unsigned i=1; i<fis->size(); i++){
    ci+=2;
    aiger_add_and(con1, ci, prev, fis->at(i));
    aiger_add_and(con4, ci, prev, fis->at(i));
    prev = ci;
  }

  aiger_add_output(con1, ci + 1, "");
  aiger_add_output(con4, ci + 1, "");
  }
  if(output_name1){
    aiger_open_and_write_to_file (con1, output_name1);
  }else{
    aiger_write_to_file (con1, aiger_ascii_mode, stdout);
  }


  /* second condition: p'(L) & !p(L) */
  /* adding kw inputs */
  for(unsigned i=0; i<km_num_inputs; i++){
    aiger_add_input(con2, (i+1)*2, "");
  }

  /* adding kw latches */
  for(unsigned i=0; i<km_num_latches; i++){
    aiger_add_input(con2, km_num_inputs*2+(i+1)*2, "");
  }

  /* adding kw ands */
  for(unsigned i=0; i<km_num_ands; i++){
    aiger_and* current_and = km_ands + i;
    aiger_add_and(con2, current_and->lhs, current_and->rhs0, current_and->rhs1);
  }
  /* adding original property p with mapping */
  for(unsigned i=0; i<m_num_ands; i++){
    aiger_and* current_and = m_ands + i;
    ci = map_kwm_to_con2(current_and->lhs);
    aiger_add_and(con2, ci, map_kwm_to_con2(current_and->rhs0), map_kwm_to_con2(current_and->rhs1));
  }
  ci += 2;
  aiger_symbol* o_s = kw_model->outputs;
  unsigned kw_output = o_s->lit;
  unsigned kw_p = neg(kw_output);

  o_s = model->outputs;
  unsigned original_output = o_s->lit;
  unsigned original_p = map_kwm_to_con2(neg(original_output));

  aiger_add_and(con2, ci, kw_p, neg(original_p));

  aiger_add_output(con2, ci, "");
  if(output_name2){
    aiger_open_and_write_to_file (con2, output_name2);
  }else{
    aiger_write_to_file (con2, aiger_ascii_mode, stdout);
  }

  /* Thrid conditioon: R(L) & R'(L) */
  
  /* adding kw inputs */
  for(unsigned i=0; i<km_num_inputs; i++){
    string s = "2 x"+to_string((i+1)*2);
    aiger_add_input(con3, (i+1)*2, s.c_str());
  }
  /* adding kw latches */
  for(unsigned i=0; i<km_num_latches; i++){
    string s;
    if(i>=(k-1)*m_num_inputs && i<(k-1)*m_num_inputs + m_num_latches){//original latches
      s = "2 x"+to_string(km_num_inputs*2+(i+1)*2);
    }else{
      s = "3 x"+to_string(km_num_inputs*2+(i+1)*2);
    }
    aiger_add_input(con3, km_num_inputs*2+(i+1)*2, s.c_str());
  }
  /* adding kw ands */
  for(unsigned i=0; i<km_num_ands; i++){
    aiger_and* current_and = km_ands + i;
    aiger_add_and(con3, current_and->lhs, current_and->rhs0, current_and->rhs1);
  }
  /* adding R(L) */
  unsigned current_index = km_total_num_literals * 2 + 2;
  vector<unsigned>* rli = new vector<unsigned>;
  for(unsigned i=0; i<m_num_latches; i++){
    aiger_symbol* latch = m_latches + i;
    unsigned l = latch->lit;
    unsigned kw_l = l + (k-1) * m_num_inputs * 2;
    unsigned reset = latch->reset;
    aiger_add_and(con3, current_index, kw_l, neg(reset));
    current_index += 2;
    aiger_add_and(con3, current_index, kw_l+1, reset);
    current_index += 2;
    aiger_add_and(con3, current_index, current_index-3, current_index-1);
    rli->push_back(current_index);
    current_index += 2;
  }

  
  if(rli->size()>0){
    prev = rli->at(0);
    for(unsigned i=1; i<rli->size(); i++){
      aiger_add_and(con3, current_index, prev, rli->at(i));
      prev = current_index;
      current_index += 2;
    }
  }
  
  if(m_num_latches == 0){
    
    aiger_add_output(con3, 1, "");
  }else{

  unsigned rl = current_index - 2;
  cout<<"rrl: "<<rl<<endl;

  vector<unsigned>* rpli = new vector<unsigned>;
  /* adding R'(L) */
  for(unsigned i=0; i<km_num_latches; i++) {
    aiger_symbol* latch = km_latches + i;//+ (k-1) * m_num_inputs;
    unsigned l = latch->lit;
    unsigned reset = latch->reset;
    aiger_add_and(con3, current_index, l, neg(reset));
    current_index += 2;
    aiger_add_and(con3, current_index, l+1, reset);
    current_index += 2;
    aiger_add_and(con3, current_index, current_index-3, current_index-1);
    rpli->push_back(current_index);
    current_index += 2;
    
  }

  prev = rpli->at(0);
  for(unsigned i=1; i<rpli->size(); i++){
    aiger_add_and(con3, current_index, prev, rpli->at(i));
    prev = current_index;
    current_index += 2;
  }

  unsigned rpl = current_index-2;
  aiger_add_and(con3, current_index, rl, neg(rpl));

      aiger_add_output(con3, (current_index), "");
  }
  if(output_name3){
    //aiger_open_and_write_to_file (con3, output_name3);
    aiger_open_and_write_to_file (con4, output_name3);
  }else{
    //aiger_write_to_file (con3, aiger_ascii_mode, stdout);
    aiger_write_to_file (con4, aiger_ascii_mode, stdout);
  }
  return 0;
}
