#!/usr/bin/env python3

import re
import sys
import os
import signal
from enum import Enum
from subprocess import Popen, PIPE
from pyparsing import nestedExpr
import pyparsing as pp
import re
import time
import uuid


def generate_smv_file(num_bits):
    f = open(sys.path[0] + "/temp_smv/q_"+str(num_bits)+".smv", "w")
    f.write("MODULE main\n")
    f.write(" VAR\n")
    f.write("  in: boolean;\n");
    for i in range(num_bits):
        f.write("    " + "a_" + str(i) +": boolean;\n")
        f.write("    " + "b_" + str(i) + ": boolean;\n")
    f.write(" DEFINE\n")
    f.write("   bad:= a_"+str(num_bits-1) + "!= b_0;\n")
    f.write(" ASSIGN\n")
    for i in range(num_bits):
        f.write("  init(a_"+str(i)+") := 0;\n")
        f.write("  init(b_"+str(i)+") := 0;\n")
    f.write("  next(a_0) := in;\n")

    for i in range(1,num_bits):
        f.write("   next(a_"+str(i) + ") := a_"+ str(i-1)+ ";\n")
    f.write("   next(b_"+ str(num_bits-1)+") := in;\n")
    for i in range(0,num_bits-1):
        f.write("   next(b_"+str(i)+") := b_"+str(i+1)+";\n")
    f.write("SPEC AG !bad");
    f.close()
    return f
    
def smv_to_aig(f, num_bits):
    print("converting smv to aig")
    exe = sys.path[0] + "/../aiger/smvtoaig "
    #outfile = " q_"+str(num_bits)+".aag"
    cmd = exe + " -s -a "  + f.name
    print("cmd: ", cmd)

    signal.signal(signal.SIGPIPE, signal.SIG_IGN)
    p=Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]
    output = output.splitlines()

    #convert_to_format1(f)
    newfile = open(sys.path[0] + "/../temp/q_"+ str(num_bits)+".aag", "w")
    for i,line in enumerate(output):
        line = line.decode('utf8')
        if i==0:
            words = line.split(" ")[:-1]
            newline = ""
            print(words)
            for j,w in enumerate(words):
                if j==4:
                    newline = newline + "1 "
                elif j==len(words)-1:
                    newline = newline + (w) + "\n"
                else:
                    newline = newline + (w + " ")
            newfile.write(newline)
        else:
            newfile.write(line+"\n")
    newfile.close()
    return newfile

def mcaiger(f, num_bits):
    print("calling Mcaiger...")
    exe = sys.path[0] + "/../mcaiger/mcaiger -n "
    cmd = exe + f.name
    print("cmd: ", cmd)
    
    time1 = time.time()
    signal.signal(signal.SIGPIPE, signal.SIG_IGN)
    p=Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]
    
    time_taken = time.time()-time1
    output = output.splitlines()
    k = output[-1].decode('utf8').split("=")[-1]
    
    print("k: ", k)
    print("McAiger computation time: ", time_taken)
    assert(int(k) == num_bits)

def aigcertify(f,k_val):
    filename = "temp/" + f.name.split("/")[-1]
    cmd  = "python3 " +  sys.path[0] + "/../py-aigcertify.py " + " " + filename + " -kis "

    print("cmd: ", cmd)
    signal.signal(signal.SIGPIPE, signal.SIG_IGN)
    p=Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]
    output = output.splitlines()
    result = output[-1].decode('utf8')
    assert(result == "SUCCESS")

def iter_aigcertify(max_num_bits):
    for i in range(2, max_num_bits):
        f = generate_smv_file(i)
        aig_file = smv_to_aig(f, i)
        mcaiger(aig_file, i)
        aigcertify(aig_file, i)

############ main #############

#max_num_bits = int(sys.argv[1])
file = sys.path[0] + "/../" + sys.argv[1]
filenames = open(file, "r")
for line in filenames:
    aigcertify(line, 0)

if(len(sys.argv)==3):
    #does not iterate
    #f = generate_smv_file(max_num_bits)
    aig_file = smv_to_aig(f, max_num_bits)
    #mcaiger(aig_file, max_num_bits)
    aigcertify(aig_file, max_num_bits)
    #iter_aigcertify(max_num_bits)
